/*  Copyright (C) 1996  Byron Stanoszek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/

/* mareport.c */

#include <sys/types.h>
#include <stdlib.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

struct io_queue {
  char *text;
  struct io_queue *next;
};

struct descriptor_data {
  struct descriptor_data *next;

  unsigned int descriptor:8;		/* file descriptor */
  int concid;				/* remote connection id */
  struct io_queue *io[2], *io_last[2];	/* text queues: in/out */
  char *raw_input;			/* current command input */
  char *addr;				/* hostname */
};

void init_args();
void serverport();
void new_connection();
struct descriptor_data *initializesock();
void init_connect();
void shutdownsock();
void free_io();
void output();
void connect_message();
void process_commands();
void process_mudcmd();
int make_socket();
void send_mud();
int mud_write();
int mud_recv();
int process_output();
int process_input();
int setsockopt();
int select();
int accept();
int shutdown();
int socket();
int bind();
int listen();
void bzero();
int setpgrp();
int connect();
int gethostname();
int getdomainname();
int getpeername();
int getsockname();
char *get_ident();
void mud_connect();
char *inet_ntoa();
unsigned long inet_addr();

int sock=-1;
int remotesock=-1;
int maxd=0;
int online=0;
int avail_descriptors=256;
int numusers=0;

long now=0L;
long lastping=0L;
long lastcheck=0L;

char hostname[60];
char domainname[60];

struct descriptor_data *descriptor_list=NULL;
struct io_queue *mudoutput=NULL;
struct io_queue *mudoutput_last=NULL;

int mudport=7348;
int remoteport=7348;
char *remotehost=NULL;
char *mudname=NULL;
char *mudinput=NULL;
char *lastmsg=NULL;

extern int errno;

#define ANSWERING_MACHINE    "login.txt"
#define TOO_MANY_CONNECTIONS "toomany.txt"
#define SERVER_PASSWORD      "password.txt"

#define GRDL_VER "v1.2"


void main(argc, argv)
int argc;
char *argv[];
{
  /* process command line arguments */
  init_args(argc, argv);

  printf("%s online at port %d.\n", remotehost?"Concentrator service":
         "Answering machine", mudport);
  fflush(stdout);
  fclose(stdout);
  fclose(stdin);
  if(fork() != 0)
    exit(0);
  setpgrp();
  now=time(NULL);

  signal(SIGPIPE, SIG_IGN);
  signal(SIGTERM, SIG_IGN);
  signal(SIGHUP, SIG_IGN);

  gethostname(hostname, 60);
  getdomainname(domainname, 60);

  /* main engine call */
  serverport();

  fprintf(stderr, "%s: Could not open port %d\n", argv[0], mudport);
  fflush(stderr);
  exit(1);
}

void init_args(argc, argv)
int argc;
char *argv[];
{
  char *cmd=argv[0];
  int a, b=0, new, opt=0;

  while(argc > 1) {
    --argc; ++argv;
    if(opt) {
      if(opt & 1) {
        /* set local port number */
        new=atoi(*argv);
        if(new < 1 || new > 9999) {
          fprintf(stderr, "%s: Invalid port number\n", cmd);
          fflush(stderr);
          exit(1);
        } else
          mudport=new;
      }
      opt=0;
      continue;
    }

    if(**argv == '-') {
      for(a=1;(*argv)[a];a++) {
        switch((*argv)[a]) {
        case 'p':
          opt|=1;
          break;
        case 'v':
          fprintf(stdout, "%s: GRDL Version Reference %s\n", cmd, GRDL_VER);
          fflush(stdout);
          exit(0);
        default:
          fprintf(stderr, "%s: illegal option -- %c\n", cmd, (*argv)[a]);
          fprintf(stderr, "Usage: %s [-v] [-p port] hostname port\n", cmd);
          fflush(stderr);
          exit(1);
        }
      }
      continue;
    }

    /* set destination hostname & port */
    if(!b++) {
      remotehost=*argv;
      continue;
    }

    new=atoi(*argv);
    if(new < 0 || new > 9999) {
      fprintf(stderr, "%s: Invalid destination port number\n", cmd);
      fflush(stderr);
      exit(1);
    } else if(new)
      remoteport=new;
  }
}

/********************/
/* Main Server Code */
/********************/

void serverport()
{
  struct descriptor_data *d, *dnext;
  struct timeval timeout;
  fd_set input_set, output_set;
  long lasttry=now-61L;
  char buf[100];
  int found=0;

  /* open local port */
  if((sock = make_socket(mudport)) == -1)
    return;
  if(sock >= maxd)
    maxd = sock+1;
  fclose(stderr);

  /* open remotehost link */
  if(remotehost)
    mud_connect();

  avail_descriptors = sysconf(_SC_OPEN_MAX) - 6;

  /* main server sequence */
  while(1) {
    process_commands();

    /* retry to connect to the mud */
    if(remotehost && remotesock == -1 && now-lasttry > 60L) {
      lasttry=now;
      mud_connect();
    }

    /* notify users if the mud goes idle every 5 minutes */
    if(remotesock != -1 && now-lastping > 299L && (now-lastping)/300L !=
       lastcheck) {
      lastcheck=(now-lastping)/300L;
      sprintf(buf, "%% The connection to %s has been idle for %ld minutes.\n",
              mudname?mudname:"the mare", (now-lastping)/60L);
      for(d=descriptor_list;d;d=d->next)
        output(d, buf);
      if(lastcheck == 6L) {
        for(d=descriptor_list;d;d=dnext) {
          dnext=d->next;
          output(d, "Please try again later.\n");
          shutdownsock(d);
        }
        shutdown(remotesock, 2);
        close(remotesock);
        remotesock=-1;
      }
    }
    
    /* clear input/output file descriptor sets */
    FD_ZERO(&input_set);
    FD_ZERO(&output_set);
    timeout.tv_sec = 60;  /* 60s time delay for select() */
    timeout.tv_usec = 0;
    
    /* set which file descriptors are ready to read input or send output */
    for(numusers=0,d=descriptor_list;d;d=d->next,numusers++) {
      FD_SET(d->descriptor, &input_set);
      if(d->io[1])
        FD_SET(d->descriptor, &output_set);
    }

    /* if we have a socket open and we have not yet exceeded the maximum */
    /* number of users, make select() look for any new connections present. */
    if(sock >= 0 && numusers < avail_descriptors)
      FD_SET(sock, &input_set);
    if(remotesock >= 0)
      FD_SET(remotesock, &input_set);
    if(remotesock >= 0 && mudoutput)
      FD_SET(remotesock, &output_set);
    
    found = select(maxd, &input_set, &output_set, (fd_set *)NULL, &timeout);
    now = time(NULL);

    if(found <= 0)
      continue;

    /* there are descriptors ready. go into input/output processing */

    /* check for a new login */
    if(sock >= 0 && FD_ISSET(sock, &input_set))
      new_connection(sock);

    /* send output to remote mud */
    if(remotesock >= 0 && FD_ISSET(remotesock, &output_set) && !mud_write()) {
      for(d=descriptor_list;d;d=dnext) {
        dnext=d->next;
        shutdownsock(d);
      }
      continue;
    }

    /* receive input from remote mud */
    if(remotesock >= 0 && FD_ISSET(remotesock, &input_set) && !mud_recv()) {
      for(d=descriptor_list;d;d=dnext) {
        dnext=d->next;
        shutdownsock(d);
      }
      continue;
    }

    /* process all output */
    for(d=descriptor_list;d;d=dnext) {
      dnext=d->next;
      if(FD_ISSET(d->descriptor, &output_set) && !process_output(d))
        shutdownsock(d);
    }

    /* process all input */
    for(d=descriptor_list;d;d=dnext) {
      dnext=d->next;
      if(FD_ISSET(d->descriptor, &input_set) && !process_input(d))
        shutdownsock(d);
    }
  }
}

int make_socket(port)
int port;
{
  struct sockaddr_in server;
  int s, opt=1;
  
  if((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    return -1;
  if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(opt)) < 0) {
    close(s);
    return -1;
  }

  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(port);
  if(bind(s, (struct sockaddr *)&server, sizeof(server))) {
    close(s);
    return -1;
  }

  listen(s, 5);
  return s;
}

void new_connection(sock)
int sock;
{
  struct sockaddr_in addr;
  struct hostent *hent;
  struct descriptor_data *d;
  char buff[100], address[100], buf[256];
  int newsock, addr_len;

  /* find incoming address */
  addr_len = sizeof(struct sockaddr);
  if((newsock = accept(sock, (struct sockaddr *)&addr, &addr_len)) < 0)
    return;

  strcpy(buff, get_ident(newsock));
  if(*buff)
    strcat(buff, "@");
  hent = gethostbyaddr((char *)&addr.sin_addr.s_addr,
                       sizeof(addr.sin_addr.s_addr), AF_INET);
  strcpy(address, hent?hent->h_name:inet_ntoa(addr.sin_addr.s_addr));
  if(!strcmp(address, "localhost")) {
    strcat(buff, hostname);
    if(*domainname) {
      strcat(buff, ".");
      strcat(buff, domainname);
    }
  } else
    strcat(buff, address);

  d=initializesock(newsock);

  /* save hostname */
  d->addr=malloc(strlen(buff)+1);
  strcpy(d->addr, buff);

  ++numusers;
  if(remotesock == -1 || now-lastping > 300L || !lastping ||
     numusers >= avail_descriptors)
    init_connect(d);

  if(remotesock != -1) {
    sprintf(buf, "REMOTECONNECT CREATE %d %s\n", newsock, buff);
    send_mud(buf);
  }
}

struct descriptor_data *initializesock(s)
int s;
{
  struct linger ling;
  struct descriptor_data *d;
  int a;

  /* initialize the new connection */

  d=malloc(sizeof(struct descriptor_data));
  fcntl(s, F_SETFL, FNDELAY);
  ling.l_onoff = 0;
  ling.l_linger = 0;
  setsockopt(s, SOL_SOCKET, SO_LINGER, (char *)&ling, sizeof(ling));
  if(s >= maxd)
    maxd=s+1;

  /* initialize data on descriptor pointer */
  d->descriptor=s;
  d->concid=0;
  d->addr=NULL;

  /* input/output strings */
  for(a=0;a<2;a++)
    d->io_last[a]=d->io[a]=NULL;
  d->raw_input=NULL;

  d->next=descriptor_list;
  descriptor_list=d;
  return d;
}

void init_connect(d)
struct descriptor_data *d;
{
  char buf[100];

  /* display login information */
  sprintf(buf, "\nWelcome to network system: %s.%s %d.\n",
          hostname, domainname, mudport);
  output(d, buf);
  sprintf(buf, "Local time is %s", ctime(&now));
  output(d, buf);
  sprintf(buf, "There are %d/%d users online.\n\n", numusers,
          avail_descriptors);
  output(d, buf);

  if(numusers >= avail_descriptors) {
    connect_message(d, TOO_MANY_CONNECTIONS);
    shutdownsock(d);
    return;
  }

  if(!lastping) {
    connect_message(d, ANSWERING_MACHINE);
    shutdownsock(d);
    return;
  }

  if(remotesock == -1)
    sprintf(buf, "%s has been down for %ld minute%s.\n", mudname?mudname:
            "The mare", (now-lastping)/60L, (now-lastping)/60L==1?"":"s");
  else
    sprintf(buf, "The connection to %s has been idle for %ld minutes.\n",
            mudname?mudname:"the mare", (now-lastping)/60L);
  output(d, buf);
  output(d, "Please wait...\n\n");
}


int process_input(d)
struct descriptor_data *d;
{
  struct io_queue *new;
  unsigned char buf[1024], temp[8192];
  int a, b, len;

  if((len = read(d->descriptor, buf, 1000)) <= 0)
    return 0;

  if(!d->raw_input) {
    *temp=0;
    b=0;
  } else {
    strcpy(temp, d->raw_input);
    b=strlen(temp);
  }

  /* filter input */
  for(a=0;a < len;a++) {
    if(buf[a] == 255 && buf[a+1] == 253 && buf[a+2] == 25) {  /* tinyfugue */
      sprintf(buf, "%d TFSENSE\n", d->concid);
      send_mud(buf);
      a+=2;
      continue;
    }
    if(buf[a] == 8 || buf[a] == 127) {  /* backspace */
      if(b > 0)
        b--;
      continue;
    }
    if(buf[a] == 10) {  /* newline */
      /* save command to input queue */
      temp[b]=0;
      new=malloc(sizeof(struct io_queue));
      new->text=malloc(b+1);
      strcpy(new->text, temp);
      if(d->io_last[0])
        d->io_last[0]->next=new;
      else
        d->io[0]=new;
      d->io_last[0]=new;
      new->next=NULL;

      *temp=0;
      b=0;
      continue;
    }
    if(buf[a] < 32 || buf[a] > 127 || b > 7999)
      continue;
    temp[b++]=buf[a];
  }
  temp[b]=0;

  /* save input for next read */
  if(d->raw_input)
    free(d->raw_input);
  d->raw_input=malloc(b+1);
  strcpy(d->raw_input, temp);

  return len;
}

int process_output(d)
struct descriptor_data *d;
{
  struct io_queue *ptr, *pnext;
  int len, amt;
  
  ptr=d->io[1];
  while(ptr) {
    pnext=ptr->next;
    if((amt=strlen(ptr->text))) {
      if((len=write(d->descriptor, ptr->text, amt)) <= 0) {
        if(!len || errno == EWOULDBLOCK)
          return -1;
        return 0;
      }
    }
    free(ptr->text);
    free(ptr);
    d->io[1]=ptr=pnext;
  }
  d->io_last[1]=NULL;
  return 1;
}

/* close remote-user socket connection */
void shutdownsock(d)
struct descriptor_data *d;
{
  struct descriptor_data *ptr;
  char buf[48];

  process_output(d);
  if(d->concid && remotesock != -1) {
    sprintf(buf, "REMOTECONNECT DELETE %d\n", d->concid);
    send_mud(buf);
  }

  shutdown(d->descriptor, 2);
  close(d->descriptor);

  free_io(d);
  for(ptr=descriptor_list;ptr;ptr=ptr->next)
    if(ptr->next == d)
      break;
  if(ptr)
    ptr->next=d->next;
  else
    descriptor_list=d->next;
  free(d);
}

void free_io(d)
struct descriptor_data *d;
{
  struct io_queue *ptr, *next;
  int a;

  for(a=0;a<2;a++) {
    for(ptr=d->io[a];ptr;ptr=next) {
      next=ptr->next;
      free(ptr->text);
      free(ptr);
    }
    d->io_last[a]=d->io[a]=NULL;
  }

  if(d->raw_input)
    free(d->raw_input);
  d->raw_input=NULL;

  if(d->addr)
    free(d->addr);
  d->addr=NULL;
}

/* adds <msg> to the output queue for a particular descriptor */
void output(d, msg)
struct descriptor_data *d;
char *msg;
{
  struct io_queue *new;

  new=malloc(sizeof(struct io_queue));
  new->text=malloc(strlen(msg)+1);
  strcpy(new->text, msg);
  if(d->io_last[1])
    d->io_last[1]->next=new;
  else
    d->io[1]=new;
  d->io_last[1]=new;
  new->next=NULL;
}

void send_mud(msg)
char *msg;
{
  struct io_queue *new;

  new=malloc(sizeof(struct io_queue));
  new->text=malloc(strlen(msg)+1);
  strcpy(new->text, msg);
  if(mudoutput_last)
    mudoutput_last->next=new;
  else
    mudoutput=new;
  mudoutput_last=new;
  new->next=NULL;
}

/* display local unix file to connection */
void connect_message(d, filename)
struct descriptor_data *d;
char *filename;
{
  FILE *fd;
  char buf[8192];

  if((fd = fopen(filename,"r")) == NULL)
    return;

  while(fgets(buf, 7999, fd))
    output(d, buf);
  fclose(fd);
  output(d, "\n");
}

void process_commands()
{
  struct descriptor_data *d;
  struct io_queue *ptr;
  int nprocessed=1;
  char buf[8192];

  while(nprocessed > 0) {
    nprocessed=0;
    for(d=descriptor_list;d;d=d->next) {
      if(!d->io[0] || !d->concid)
        continue;

      /* send data to remotesock */
      sprintf(buf, "%d %s\n", d->concid, d->io[0]->text);
      send_mud(buf);

      ptr=d->io[0];
      if(!(d->io[0]=ptr->next))
        d->io_last[0]=NULL;
      free(ptr->text);
      free(ptr);
      nprocessed++;
    }
  }
}

int mud_write()
{
  struct io_queue *ptr, *pnext;
  
  ptr=mudoutput;
  while(ptr) {
    pnext=ptr->next;
    if(write(remotesock, ptr->text, strlen(ptr->text)) < 0) {
      shutdown(remotesock, 2);
      close(remotesock);
      remotesock=-1;
      return 0;
    }
    free(ptr->text);
    free(ptr);
    mudoutput=ptr=pnext;
  }
  mudoutput_last=NULL;
  return 1;
}

int mud_recv()
{
  unsigned char buf[1024], temp[8192];
  int a, b, len;

  if((len = read(remotesock, buf, 1000)) < 0) {
    shutdown(remotesock, 2);
    close(remotesock);
    remotesock=-1;
    return 0;
  }
  if(!len)
    return 1;

  if(!mudinput) {
    *temp=0;
    b=0;
  } else {
    strcpy(temp, mudinput);
    b=strlen(temp);
  }

  /* filter input */
  for(a=0;a < len;a++) {
    if(buf[a] == 10) {  /* newline */
      /* save command to input queue */
      temp[b]=0;
      process_mudcmd(temp);
      *temp=0;
      b=0;
      continue;
    }
    if(b > 8100)
      continue;
    temp[b++]=buf[a];
  }
  temp[b]=0;

  /* save input for next read */
  if(mudinput)
    free(mudinput);
  mudinput=malloc(b+1);
  strcpy(mudinput, temp);

  return len;
}

void process_mudcmd(cmd)
char *cmd;
{
  struct descriptor_data *d;
  char buf[256], *p, *s;
  int a;

  if(!strcmp(cmd, "PING")) {
    lastping=now;
    lastcheck=0L;
    return;
  }
  if(!strncmp(cmd, "IDENT ", 6)) {
    lastping=now;
    lastcheck=0L;
    if(mudname)
      free(mudname);
    mudname=malloc(strlen(cmd+6)+1);
    strcpy(mudname, cmd+6);
    for(d=descriptor_list;d;d=d->next)
      if(!d->concid) {
        sprintf(buf, "REMOTECONNECT CREATE %d %s\n", d->descriptor, d->addr);
        send_mud(buf);
      }
    return;
  }
  if(!strncmp(cmd, "CLOSE ", 6)) {
    for(d=descriptor_list;d;d=d->next)
      if(d->concid == atoi(cmd+6)) {
        shutdownsock(d);
        return;
      }
    return;
  }
  if(!strncmp(cmd, "OPEN ", 5)) {
    for(d=descriptor_list;d;d=d->next)
      if(d->descriptor == atoi(cmd+5)) {
        if(!(s=strchr(cmd+5,' ')))
          return;
        if(d->concid) {
          sprintf(buf, "REMOTECONNECT DELETE %d\n", atoi(s+1));
          send_mud(buf);
        } else
          d->concid=atoi(s+1);
        return;
      }
    return;
  }

  if(!(a=atoi(cmd)))
    return;
  if(!(s=strchr(cmd, ' '))) {
    if(!lastmsg)
      return;
    s=lastmsg;
  } else {
    s++;
    while((p=strchr(s,'\r')))
      *p='\n';
    if(lastmsg)
      free(lastmsg);
    lastmsg=malloc(strlen(s)+1);
    strcpy(lastmsg, s);
  }

  for(d=descriptor_list;d;d=d->next)
    if(d->concid == a) {
      output(d, s);
      return;
    }
}


char *parse_up(str,delimit)
char **str;
int delimit;
{
  int deep=0;
  char *s=*str, *os=*str;

  if (!*s)
    return 0;
  while(*s && (*s!=delimit))
    if (*s++=='{') {
      deep=1;
      while(deep && *s)
      switch(*s++) {
      case '{':
        deep++;
        break;
      case '}':
        deep--;
        break;
      }
    }

  if (*s)
    *s++=0;
  *str=s;
  return(os);
}

/* query the remote host's identity */
char *get_ident(sock)
int sock;
{
  fd_set input_set, output_set;
  struct sockaddr_in localaddr, remoteaddr, sin_laddr, sin_faddr;
  struct timeval timeout;
  struct linger ling;
  int local, remote;
  int fd, len, res, opt=1;
  char buf[100], *str=NULL, *r, *s;
  static char rslt[32];

  /* set up local/remote addresses */
  len = sizeof(remoteaddr);
  if(getpeername(sock, (struct sockaddr *) &remoteaddr, &len) < 0)
    return "";
  len = sizeof(localaddr);
  if(getsockname(sock, (struct sockaddr *) &localaddr, &len) < 0)
    return "";
  local=ntohs(localaddr.sin_port);
  remote=ntohs(remoteaddr.sin_port);

  /* open socket connection to remote address */
  if((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    return "";
  ling.l_onoff = 0;
  ling.l_linger = 0;
  setsockopt(fd, SOL_SOCKET, SO_LINGER, (void *) &ling, sizeof(ling));
  setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *) &opt, sizeof(opt));

  bzero((char *)&sin_laddr, sizeof(sin_laddr));
  sin_laddr.sin_family = AF_INET;
  sin_laddr.sin_addr = localaddr.sin_addr;
  sin_laddr.sin_port = 0;
  if (bind(fd, (struct sockaddr *) &sin_laddr, sizeof(sin_laddr))) {
    close(fd);
    return "";
  }

  sin_faddr.sin_family = AF_INET;
  sin_faddr.sin_addr = remoteaddr.sin_addr;
  sin_faddr.sin_port = htons(113); /* identd port */
  if(connect(fd, (struct sockaddr *) &sin_faddr, sizeof(sin_faddr)) < 0) {
    close(fd);
    return "";
  }

  FD_ZERO(&input_set);
  FD_ZERO(&output_set);
  FD_SET(fd, &input_set);
  FD_SET(fd, &output_set);
  timeout.tv_sec=5;
  timeout.tv_usec=0;

  if((res = select(FD_SETSIZE, &input_set, &output_set,
     (fd_set *)NULL, &timeout)) < 0) {
    close(fd);
    return "";
  } if(!res || (!FD_ISSET(fd, &input_set) && !FD_ISSET(fd, &output_set))) {
    close(fd);
    return "";
  }

  /* send output through socket requesting identification */
  FD_ZERO(&output_set);
  FD_SET(fd, &output_set);

  if ((res = select(FD_SETSIZE, (fd_set *)NULL, &output_set,
      (fd_set *)NULL, &timeout)) <= 0) {
    close(fd);
    return "";
  }
  sprintf(buf, "%d , %d\r\n", remote, local);
  write(fd, buf, strlen(buf));

  /* wait for return input */
  FD_ZERO(&input_set);
  FD_SET(fd, &input_set);

  if ((res = select(FD_SETSIZE, &input_set, (fd_set *)0,
      (fd_set *)0, &timeout)) <= 0) {
    close(fd);
    return "";
  }

  /* read input */
  len=0;
  while(len < 99 && (res = read(fd, buf+len, 1)) == 1 &&
        buf[len] != '\n' && buf[len] != '\r')
    len++;
  if(res <= 0) {
    close(fd);
    return "";
  }

  /* parse input string and search for userid */
  buf[len]=0;
  s=buf;
  res=0;
  while((r=parse_up(&s,':'))) {
    res++;
    str=r;
  }
  if(res != 4) {
    close(fd);
    return "";
  }

  if(*str == ' ')
    str++;
  strncpy(rslt, str, 31);
  rslt[31]=0;

  close(fd);
  return rslt;
}

void mud_connect()
{
  FILE *fd;
  struct sockaddr_in addr;
  struct hostent *hp=NULL;
  char address[100], buf[100], pass[25];
  char *p;

  strncpy(address, remotehost, 99);
  address[99]=0;
  p = address;
  while (*p && ((*p == '.') || isdigit(*p))) p++;
  if(*p) {
    hp=gethostbyname(address);
    if(!hp)
      exit(2);
    memcpy((char *)&addr.sin_addr,hp->h_addr,hp->h_length);
  } else {
    u_long f;
    if((f = inet_addr(address)) == -1L)
      exit(2);
    memcpy((char *)&addr.sin_addr,(char *)&f,sizeof(f));
  }

  addr.sin_port = htons(remoteport);
  addr.sin_family = AF_INET;
  if((remotesock = socket(AF_INET,SOCK_STREAM,0)) < 0)
    exit(2);
  if(connect(remotesock,(struct sockaddr *)&addr,sizeof(addr)) < 0) {
    close(remotesock);
    remotesock=-1;
    return;
  }
  if(remotesock >= maxd)
    maxd=remotesock+1;

  /* read password and send the initialization message */
  if((fd=fopen(SERVER_PASSWORD,"r"))) {
    fgets(pass, 24, fd);
    fclose(fd);
    sprintf(buf, "REMOTECONNECT NEW %s", pass);
    send_mud(buf);
  } else
    send_mud("REMOTECONNECT NEW rukzuk\n");
}
